#include <linux/init.h>
#include <linux/sched.h>
#include <linux/kernel.h> 
#include <linux/errno.h>      
#include <linux/timer.h>
#include <linux/types.h>        
#include <linux/kdev_t.h>
#include <linux/vmalloc.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <asm/io.h>
#include <asm/arch/hardware.h>
#include <asm/atomic.h>
#include <asm/semaphore.h>
#include <asm/current.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>

#include "ts73xx_dma.h"

struct ts73x0dma_dev {
	volatile unsigned long dmaregs;
	volatile unsigned long sysconfpga;
	volatile unsigned short *fpgaregs;
	struct semaphore sem;
	wait_queue_head_t qdma;
	int irq;
};

static int buf_size = 4096;	//1 page
static dma_addr_t dmaaddr;
static u_char *dmabuf;
//DECLARE_COMPLETION(dma_completion);

static struct ts73x0dma_dev *ts73x0dma = NULL;
/*
static irqreturn_t ts73x0_int_handler(int irq, void *vidp)
{
	struct ts73x0dma_dev *dev = (struct ts73x0dma_dev *)vidp;
	DMA_WRITE(INTERRUPT, 0x0);
	complete(&dma_completion);
	BUG_ON(irq != dev->irq);
	return IRQ_HANDLED;
}
*/
int dma_op (unsigned long dev_addr, unsigned long v_addr, unsigned int len, short rw) {
/*
	if (down_interruptible(&ts73x0dma->sem)) {
		return -ERESTARTSYS;
	}
*/

	DMA_WRITE(CONTROL, 0x01085200); DMA_READ(CONTROL);
	ts73x0dma->fpgaregs[12] = dev_addr; //select NAND/SD

	if(rw == MTD_READ) {
		ts73x0dma->fpgaregs[8] = (len/2)|(1<<12);
		DMA_WRITE(INTERRUPT, DMA_READ(INTERRUPT) | (1<<1)); 
		DMA_WRITE(DAR_BASE0, dmaaddr);	//phys address
		DMA_WRITE(SAR_BASE0, 0x600ff020);
		DMA_WRITE(BCR0, len);
		DMA_WRITE(CONTROL, 0x01085208); DMA_READ(CONTROL);

		while (!(DMA_READ(STATUS) & 0x40));

		DMA_WRITE(STATUS, 0x0);	
		DMA_WRITE(INTERRUPT, 0x0);
		memcpy((unsigned char *)v_addr, dmabuf, len);
	} else if(rw == SD_READ || rw == SD_WRITE) {
		if(rw == SD_WRITE) {
			ts73x0dma->fpgaregs[8] = (len/2);
			DMA_WRITE(INTERRUPT, DMA_READ(INTERRUPT) | (1<<1)); 
			DMA_WRITE(DAR_BASE0, 0x600ff020);
			DMA_WRITE(SAR_BASE0, dmaaddr);
		}else if (rw == SD_READ) {
    			ts73x0dma->fpgaregs[8] = (len/2)|(1<<12);
			DMA_WRITE(INTERRUPT, DMA_READ(INTERRUPT) | (1<<1)); 
    			DMA_WRITE(SAR_BASE0, 0x600ff020);
			DMA_WRITE(DAR_BASE0, dmaaddr);
  		}

		DMA_WRITE(BCR0, len);
		DMA_WRITE(CONTROL, 0x01085208);

		while (!(DMA_READ(STATUS) & 0x40));

		DMA_WRITE(STATUS, 0x0);	
		DMA_WRITE(INTERRUPT, 0x0);
		memcpy((unsigned char *)v_addr, dmabuf, len);
	}

//	up(&ts73x0dma->sem);

	return 0;
}
EXPORT_SYMBOL(dma_op);

static int __init dma_bus_init(void)
{
	struct ts73x0dma_dev *dev = NULL;
	dev = kmalloc(sizeof (struct ts73x0dma_dev), GFP_KERNEL);
	memset (dev, 0, sizeof (struct ts73x0dma_dev));
	dev->dmaregs = (unsigned long) __ioremap(DMA_M2M_1_BASE, DMA_M2M_1_SIZE, 0);
	dev->sysconfpga = (unsigned long)__ioremap(SYSCONFPGA_BASE, 0xFF, 0);
	dev->fpgaregs = (unsigned short *)__ioremap(FPGA_REGS, 0xFF, 0);
	init_MUTEX(&dev->sem);
	//init_waitqueue_head(&dev->qdma);
	//dev->irq = IRQ_EP93XX_DMAM2M1;	//DMA_M2M1 = IRQ 18
	//request_irq(dev->irq, ts73x0_int_handler, SA_INTERRUPT, NULL, dev);
	dmabuf = dma_alloc_coherent(NULL, buf_size, &dmaaddr, GFP_KERNEL);

	if(NULL == dmabuf) {
		printk("tsdma.c: unable to allocate buffer\n");
		//free_irq(dev->irq, dev);
		return -1;
	} else memset(dmabuf, 0, buf_size);

	ts73x0dma = dev;
	return 0;
}
subsys_initcall(dma_bus_init);
