/*
 *  linux/drivers/serial/tsuart-pt0100.c
 *
 *  Linux 2.6 Driver for Technologic Systems UART
 *
 *  (c) 2007 Technologic Systems
 *
 * This source code is hereby licensed to the Linux 2.4 kernel under the
 * GPL, and to Technologic Systems customers under the BSD license.
 *
 * 2008-09-24:mos:initial version
 */
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/serial_core.h>
#include <asm/uaccess.h>
#include "tsuart1.h"

//---------------------------------------------------------------------------
// TS-7800 + PT-0100 daughter board
// This is the daughter board for Mr. Foreman which talks to the IR camera
//---------------------------------------------------------------------------
#define BOARD_ID_PT100 0x0100


// mask off bits 0xFC000 in register 0xE8000038
// this sets C14-C19 on the PC104 bus to GPIO, which is needed for
// UART #8 and UART #9 signals to appear on the header.
int set104(void) {
  volatile unsigned long *tmp;
  tmp = ((unsigned long *) __ioremap(0xE8000000, 4096,0));
  if (tmp == (unsigned long *)0xE8000000) {
    printk(KERN_ERR "tsuart1: cannot map 0xE8000000\n");
    return 0;
  }
  tmp[0x30/4] = 0xAAAAAAAA;
  tmp[0x34/4] = 0xAAAAAAAA;
  return 1;
}

void (*old_shutdown)(struct uart_port *);
static struct uart_ops tsuart_ops_7800;
void shutdown_7800(struct uart_port *p) {

  // printk("Waiting for transmitter to be empty to shut down...");
  // assume we won't be able to send any more bytes while waiting
  while (!bit_TRE(TSPORT(p)->get(TSPORT(p),0))) {
    printk(".");
    set_current_state(TASK_INTERRUPTIBLE);
    schedule_timeout(1);
  }
  printk("done!\n");
  set_current_state(TASK_RUNNING);
  old_shutdown(p);
}

static int tryInitPort(int offset) {
  struct tsuart_port *ts;

  ts = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
  memset(ts,0,sizeof(struct tsuart_port));

  init_tsuart_port(ts);
  ts->u.iotype = SERIAL_IO_MEM;
  ts->get = tsuartGetMEM16;
  ts->put = tsuartPutMEM16;
  ts->mdmctrl = 0;
  ts->debugFlags = portDebug;
  ts->u.mapbase = 0xE8000080 + 4 * offset;
  ts->u.irq = 64 + 27; // to 64+25
  //ts->localStatus |= PORT_STATUS_SHIRQ;
  ts->boardId = BOARD_ID_PT100;

  if (tsuart_register_port(0,ts) != 0) {
    printk("Failed to register port!\n");
    return 0;
  }

  old_shutdown = ts->u.ops->shutdown;
  tsuart_ops_7800 = *(ts->u.ops);
  tsuart_ops_7800.shutdown = shutdown_7800;
  ts->u.ops = &tsuart_ops_7800;

  return 1;
}

//---------------------------------------------------------------------------
// KERNEL MODULE
//---------------------------------------------------------------------------
static int __init tsuart_init(void)
{
  int i,maxport;

  if (!set104()) {
    return -1;
  }
  tryInitPort(0);
  return 0;
}

static void __exit tsuart_exit(void)
{
}

module_init(tsuart_init);
module_exit(tsuart_exit);

MODULE_DESCRIPTION("Technologic Systems TS-UART driver");


